#include <alloca.h>
#include <ctype.h>
#include <errno.h>
#include <error.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/mman.h>

/* This will be running setuid root, so be careful! */

static char * safeEnviron[] = {
	"PATH=/bin:/sbin:/usr/bin:/usr/sbin",
	"HOME=/root",
	NULL
};

#define FOUND_FALSE -1
#define NOT_FOUND 0
#define FOUND_TRUE 1

static void
usage(void) {
    fprintf(stderr, "usage: usernetctl <interface-config> <up|down|report>\n");
    exit(1);
}

static void
testSafe(const char *ifaceConfig, int fd, struct stat *sb) {
    int rc;

    if (strstr (ifaceConfig, "/../"))
	error (1, EINVAL, "%s", ifaceConfig);

    rc = (fd < 0) ? stat (ifaceConfig, sb) : fstat (fd, sb);
    if (rc < 0)
	error (1, errno, "stat: %s", ifaceConfig);

    if (!S_ISREG(sb->st_mode))
	error (1, 0, "%s: not a normal file", ifaceConfig);

    if (sb->st_uid)
	error (1, errno, "%s: should be owned by root", ifaceConfig);
    
    if (sb->st_mode & S_IWOTH)
	error (1, errno, "%s: should not be world writeable", ifaceConfig);
}


static int
userCtl(char *file) {
    char *buf;
    char *contents;
    char *chptr = 0;
    char *next = 0;
    int fd = -1, retval = NOT_FOUND;
    struct stat st;

    testSafe (file, fd, &st);

    fd = open (file, O_RDONLY|O_NOCTTY|O_NONBLOCK);
    if (fd < 0)
	error (1, errno, "open: %s", file);

    testSafe (file, fd, &st);

    buf = contents = malloc (st.st_size + 2);

    if (read (fd, contents, st.st_size) != st.st_size)
	error (1, errno, "read: %s", file);
    close(fd);

    contents[st.st_size] = '\n';
    contents[st.st_size + 1] = '\0';

    /* Each pass parses a single line (until an answer is found),  The contents
       itself points to the beginning of the current line */
    while (*contents) {
	chptr = contents;
	while (*chptr != '\n') chptr++;
	next = chptr + 1;
	while (chptr >= contents && isspace(*chptr)) chptr--;
	*(++chptr) = '\0';

	if (!strncmp(contents, "USERCTL=", 8)) {
	    contents += 8;
	    if ((contents[0] == '"' &&
		 contents[strlen(contents) - 1] == '"') ||
		(contents[0] == '\'' &&
		 contents[strlen(contents) - 1] == '\''))
		{
		contents++;
		contents[strlen(contents) - 1] = '\0';
	    }

	    if (!strcmp(contents, "yes") || !strcmp(contents, "true")) 
		retval = FOUND_TRUE;
	    else 
		retval = FOUND_FALSE;

	    break;
	}

	contents = next;
    }

    free (buf);

    return retval;
}

int
main(int argc, char ** argv) {
    char * ifaceConfig;
    char * chptr;
    char * cmd = 0;
    int report = 0;
    char tmp;
    const char *ns_dir = "/etc/sysconfig/network-scripts";

    if (argc != 3) usage();

    if (!strcmp(argv[2], "up")) {
	cmd = "./ifup";
    } else if (!strcmp(argv[2], "down")) {
	cmd = "./ifdown";
    } else if (!strcmp(argv[2], "report")) {
	report = 1;
    } else {
	usage();
    }

    if (chdir(ns_dir))
	error (1, errno, "chdir: %s", ns_dir);

    /* force the interface configuration to be in the current directory */
    chptr = ifaceConfig = argv[1];
    while (*chptr) {
	if (*chptr == '/')
	    ifaceConfig = chptr + 1;
	chptr++;
    }

    /* automatically prepend "ifcfg-" if it is not specified */
    if (strncmp(ifaceConfig, "ifcfg-", 6)) {
	char *temp;

	temp = (char *) alloca(strlen(ifaceConfig) + 7);
	strcpy(temp, "ifcfg-");
	/* strcat is safe because we got the length from strlen */
	strcat(temp, ifaceConfig);
	ifaceConfig = temp;
    }
    
	if ( getuid() )
    switch (userCtl(ifaceConfig)) {
	char *dash;

	case NOT_FOUND:
	    /* a `-' will be found at least in "ifcfg-" */
	    dash = strrchr(ifaceConfig, '-');
	    if (*(dash-1) != 'g') {
		/* This was a clone configuration; ask the parent config */
		tmp = *dash;
		*dash = '\0';
		if (userCtl(ifaceConfig) == FOUND_TRUE) {
		    /* exit the switch; users are allowed to control */
		    *dash = tmp;
		    break;
		}
		*dash = tmp;
	    }
	    /* else fall through */
	case FOUND_FALSE:
	    if (! report)
	        fprintf(stderr,
			"Users are not allowed to control this interface.\n");
	    exit(1);
	    break;
    }

    /* looks good to me -- let's go for it if we are changing the interface,
     * report good status to the user otherwise */

    if (report)
	exit(0);

    /* pppd wants the real uid to be the same as the effective (god only
       knows why when it works fine setuid out of the box) */
    setuid(geteuid());

    execle(cmd, cmd, ifaceConfig, NULL, safeEnviron);
    fprintf(stderr, "exec of %s failed: %s\n", cmd, strerror(errno));
    
    exit(1);
}
